from qiskit import QuantumCircuit, transpile
from qiskit_aer import AerSimulator


def demo_function(shotsAmount=1000):
    simulator = AerSimulator()

    circuit = QuantumCircuit(2, 2)
    circuit.h(0)
    circuit.cx(0, 1)
    circuit.measure([0, 1], [0, 1])

    compiled_circuit = transpile(circuit, simulator)
    job = simulator.run(compiled_circuit, shots=shotsAmount)
    result = job.result()
    counts = result.get_counts()

    print("Total count for 00 and 11 are:", counts)
    print(circuit)


result = demo_function(2000)



import os
import unittest
from unittest.mock import patch, MagicMock
from q8s.utils import get_docker_image
from q8s.enums import Target
from q8s.project import ProjectNotFoundException, CacheNotBuiltException


class TestGetDockerImage(unittest.TestCase):

    @patch("q8s.utils.Project")
    def test_get_docker_image_success(self, MockProject):
        mock_project = MockProject.return_value
        mock_project.cached_images.return_value = "mock_image"
        image = get_docker_image(target=Target.cpu)
        self.assertEqual(image, "mock_image")

    @patch("q8s.utils.Project")
    @patch("q8s.utils.os.environ.get")
    def test_get_docker_image_project_not_found(self, mock_environ_get, MockProject):
        MockProject.side_effect = ProjectNotFoundException
        mock_environ_get.return_value = "env_image"
        image = get_docker_image(target=Target.gpu)
        self.assertEqual(image, "env_image")

    @patch("q8s.utils.Project")
    @patch("q8s.utils.os.environ.get")
    def test_get_docker_image_cache_not_built(self, mock_environ_get, MockProject):
        MockProject.return_value.cached_images.side_effect = CacheNotBuiltException
        mock_environ_get.return_value = "env_image"
        image = get_docker_image(target=Target.gpu)
        self.assertEqual(image, "env_image")

    @patch("q8s.utils.Project")
    @patch("q8s.utils.os.environ.get")
    def test_get_docker_image_general_exception(self, mock_environ_get, MockProject):
        MockProject.side_effect = Exception("General error")
        mock_environ_get.return_value = "env_image"
        image = get_docker_image(target=Target.cpu)
        self.assertEqual(image, "env_image")


if __name__ == "__main__":
    unittest.main()



from importlib.metadata import (
    packages_distributions,
    version,
)

from .code_analyzer import CodeAnalyzer

mapping = {
    "qiskit-aer": "qiskit-aer-gpu",
}


class Parser:
    def __init__(self) -> None:
        self.installed = packages_distributions()

    def mapToPip(self, package: str) -> str:
        """
        Map the package to the pip package.
        """

        try:
            return self.installed[package][0]
        except KeyError:
            return None

    def parse(self, code: str) -> str:
        """
        Parse the code and returns the requirements.
        """
        imports = CodeAnalyzer(code).getImports()

        pips = list(map(self.mapToPip, imports))

        return "\n".join(
            list(
                map(
                    lambda x: "{name}=={version}".format(
                        name=mapping[x] if x in mapping.keys() else x,
                        version=version(x),
                    ),
                    pips,
                )
            )
        )



import os
from typing import Dict
from kubernetes import client
from q8s.constants import WORKSPACE
from q8s.enums import Target
from q8s.plugins.job_template_spec import hookimpl

MEMORY = os.environ.get("MEMORY", "32Gi")


class CUDAJobTemplatePlugin:
    """
    This plugin is used to create a job template for a GPU job.
    """

    @hookimpl
    def makejob(
        self,
        name: str,
        registry_pat: str | None,
        registry_credentials_secret_name: str,
        container_image: str,
        env: Dict[
            str,
            str | None,
        ],
        target: Target,
    ) -> client.V1PodTemplateSpec:

        if target != Target.gpu:
            return None

        container = client.V1Container(
            name="quantum-routine",
            image=container_image,
            env=env,
            command=["python"],
            args=[f"{WORKSPACE}/main.py"],
            image_pull_policy="Always",
            resources=(
                client.V1ResourceRequirements(
                    limits=(
                        {
                            "cpu": "2",
                            "ephemeral-storage": "50Gi",
                            "memory": MEMORY,
                            "nvidia.com/gpu": "1",
                            # "qubernetes.dev/qpu": "1",
                        }
                    ),
                    requests=(
                        {
                            "cpu": "2",
                            "ephemeral-storage": "0",
                            "memory": MEMORY,
                            "nvidia.com/gpu": "1",
                            # "qubernetes.dev/qpu": "1",
                        }
                    ),
                )
            ),
            volume_mounts=[
                client.V1VolumeMount(
                    name="app-volume", mount_path=WORKSPACE, read_only=True
                )
            ],
        )

        template = client.V1PodTemplateSpec(
            metadata=client.V1ObjectMeta(labels={"app": name}),
            spec=client.V1PodSpec(
                containers=[container],
                image_pull_secrets=(
                    [
                        client.V1LocalObjectReference(
                            name=registry_credentials_secret_name
                        )
                    ]
                    if registry_pat
                    else []
                ),
                runtime_class_name="nvidia",
                restart_policy="Never",
                volumes=[
                    client.V1Volume(
                        name="app-volume",
                        config_map=client.V1ConfigMapVolumeSource(name=name),
                    )
                ],
            ),
        )

        return template



import unittest

from q8s_kernel.deps.parser import Parser


class TestParser(unittest.TestCase):
    def test_parse(self):
        self.assertEqual(1, 1)

        with open("tests/fixtures/qiskit.py") as f:
            code = f.read()
            parser = Parser()
            result = parser.parse(code)

            self.assertEqual(result, "qiskit==1.0.0\nqiskit-aer-gpu==0.13.3")


if __name__ == "__main__":
    unittest.main()



import base64
from json import JSONEncoder, loads
import logging
import os
import random
import string
from time import sleep
from dotenv import dotenv_values
from kubernetes import client, config, watch
import pluggy
from rich.progress import Progress

from q8s.constants import WORKSPACE
from q8s.enums import Target
from q8s.plugins.job_template_spec import JobTemplatePluginSpec
from q8s.plugins.cpu_job import CPUJobTemplatePlugin
from q8s.plugins.cuda_job import CUDAJobTemplatePlugin
from q8s.utils import extract_non_none_value


def load_env():
    env = dotenv_values(".env.q8s")

    return env


class K8sContext:
    container_image: str | None = None
    registry_pat: str | None = None
    jupyter_logger: None
    target: Target = Target.gpu
    jm: pluggy.PluginManager = pluggy.PluginManager("q8s")
    __progress: Progress | None

    def __init__(self, kubeconfig: str, logger=None, progress: Progress = None):
        """
        Initialize the Kubernetes context.
        """
        self.__progress = progress

        self.jm.add_hookspecs(JobTemplatePluginSpec)
        self.jm.register(CPUJobTemplatePlugin())
        self.jm.register(CUDAJobTemplatePlugin())

        task_config = self.__progress.add_task(
            "[cyan]Loading configuration...", total=1
        )

        config.load_kube_config(kubeconfig)
        self.__progress.console.print("Cluster configuration loaded")
        self.__progress.update(task_config, completed=True)

        try:
            _, active_context = config.list_kube_config_contexts(config_file=kubeconfig)
            self.namespace = active_context["context"]["namespace"]
        except KeyError:
            self.namespace = "default"
        self.__progress.console.print(f"Active namespace: {self.namespace}")

        self.core_api_instance = client.CoreV1Api()
        self.batch_api_instance = client.BatchV1Api()

        self.name = f"qubernetes-job-{K8sContext.get_id()}"

        self.__env = load_env()

        self.jupyter_logger = logger

    @staticmethod
    def get_id():
        """
        Generate a random ID.
        """
        return "".join(random.choices(string.ascii_lowercase + string.digits, k=6))

    def set_container_image(self, image: str):
        self.container_image = image

    def set_registry_pat(self, pat: str):
        self.registry_pat = pat

    def set_target(self, target: Target):
        self.target = target

    def create_job_object(self, code: str) -> client.V1Job:
        return None

    def __registry_credentials_secret_name(self):
        return f"{self.name}-regcred"

    def __create_job_object(self, code: str):
        """
        Create a job object with the given code.
        """
        prepare_task = self.__progress.add_task("[cyan]Prepare job...", total=1)
        env = self.__prepare_environment()

        self.jm.hook.prepare(
            target=self.target, name=self.name, namespace=self.namespace, env=self.__env
        )

        logging.debug(f"Container image: {self.container_image}")

        template = extract_non_none_value(
            self.jm.hook.makejob(
                code=code,
                env=env,
                container_image=self.container_image,
                target=self.target,
                registry_credentials_secret_name=self.__registry_credentials_secret_name(),
                name=self.name,
                registry_pat=self.registry_pat,
            )
        )

        # Create the specification of deployment
        spec = client.V1JobSpec(template=template)  # , ttl_seconds_after_finished=10

        # Instantiate the job object
        job_spec = client.V1Job(
            api_version="batch/v1",
            kind="Job",
            metadata=client.V1ObjectMeta(
                name=self.name,
                namespace=self.namespace,
                labels={"qubernetes.dev/job.type": "jupyter"},
            ),
            spec=spec,
        )

        job = self.batch_api_instance.create_namespaced_job(
            body=job_spec, namespace=self.namespace
        )
        self.__progress.console.print("Job created")

        self.__create_config_map_object(code, job)
        self.__progress.console.print("Application code created")
        self.__create_environment_secret()
        self.__progress.console.print("Environment variables created")

        if self.registry_pat:
            self.__create_registry_credentials_secret()

        self.__progress.advance(prepare_task, 1)
        return job

    def __create_config_map_object(self, code: str, job: client.V1Job):
        """
        Create a ConfigMap object with the given code.
        """
        # Configureate ConfigMap from a local file
        configmap = client.V1ConfigMap(
            api_version="v1",
            kind="ConfigMap",
            data={"main.py": code},
            metadata=client.V1ObjectMeta(
                name=self.name,
                owner_references=[
                    client.V1OwnerReference(
                        api_version="batch/v1",
                        kind="Job",
                        name=job.metadata.name,
                        uid=job.metadata.uid,
                        # block_owner_deletion=True,
                        # controller=True,
                    )
                ],
            ),
        )

        self.core_api_instance.create_namespaced_config_map(
            namespace=self.namespace, body=configmap
        )

    def __create_environment_secret(self):
        """
        Create a Secret object with the environment variables.
        """
        data = {}

        for key in self.__env.keys():
            data[key] = base64.b64encode(self.__env[key].encode()).decode()

        secret = client.V1Secret(
            api_version="v1",
            kind="Secret",
            type="Opaque",
            immutable=True,
            data=data,
            metadata=client.V1ObjectMeta(
                name=self.name,
                namespace=self.namespace,
                # owner_references=[
                #     client.V1OwnerReference(
                #         api_version="v1",
                #         kind="Job",
                #         name=job.metadata.name,
                #         uid=job.metadata.uid,
                #         # block_owner_deletion=True,
                #         # controller=True,
                #     )
                # ],
            ),
        )

        self.core_api_instance.create_namespaced_secret(
            namespace=self.namespace, body=secret
        )

    def __create_registry_credentials_secret(self):
        """
        Create a Secret object with the registry credentials.
        """
        segments = self.container_image.split("/")
        # Find user name for images on Docker Hub
        username = segments[0] if len(segments) == 2 else segments[1]
        registry = segments[0] if len(segments) == 3 else "https://index.docker.io/v1/"

        config = {
            "auths": {
                registry: {
                    "auth": base64.b64encode(
                        f"{username}:{self.registry_pat}".encode()
                    ).decode(),
                }
            }
        }

        secret = client.V1Secret(
            api_version="v1",
            kind="Secret",
            type="kubernetes.io/dockerconfigjson",
            immutable=True,
            metadata=client.V1ObjectMeta(
                name=self.__registry_credentials_secret_name(),
                namespace=self.namespace,
            ),
            data={
                ".dockerconfigjson": base64.b64encode(
                    JSONEncoder().encode(config).encode()
                ).decode(),
            },
        )

        self.core_api_instance.create_namespaced_secret(
            namespace=self.namespace, body=secret
        )

    def __delete_job(self):
        """
        Delete the job and its associated resources.
        """
        cleanup_task = self.__progress.add_task("[cyan]Cleaning up...", total=1)

        if self.registry_pat:
            self.core_api_instance.delete_namespaced_secret(
                self.__registry_credentials_secret_name(), self.namespace
            )
            self.__progress.console.print("Registry credentials removed")

        self.core_api_instance.delete_namespaced_secret(self.name, self.namespace)
        self.__progress.console.print("Environment variables removed.")

        self.core_api_instance.delete_namespaced_config_map(
            self.name,
            self.namespace,
            body=client.V1DeleteOptions(propagation_policy="Foreground"),
        )
        self.__progress.console.print("Application code removed.")

        self.jm.hook.cleanup(name=self.name, namespace=self.namespace)

        api_response = self.batch_api_instance.delete_namespaced_job(
            self.name,
            self.namespace,
            body=client.V1DeleteOptions(propagation_policy="Foreground"),
            _preload_content=False,
        )

        try:
            data = loads(api_response.data)

            self.__progress.console.print(
                f"Job removed. status='{data['status']['conditions'][0]['type']}'"
            )
        except:
            self.__progress.console.print(
                f"Job removed. status='{str(api_response.status)}'"
            )
        finally:
            self.__progress.advance(cleanup_task, 1)

    def __get_job_logs(self, name="qiskit-aer-gpu"):
        """
        Get the logs of the job.
        """
        api_response = self.core_api_instance.read_namespaced_pod_log(
            name=name, namespace=self.namespace
        )
        logging.debug("Job logs='%s'" % str(api_response))
        return api_response

    def __get_pods_in_job(self):
        """
        Get the pods in the job.
        """
        pods = self.core_api_instance.list_namespaced_pod(
            self.namespace, label_selector=f"app={self.name}"
        )

        pod_name = pods.items[0].metadata.name

        return pod_name

    def __complete_and_get_job_status(self):
        """
        Wait for the job to complete and get its status.
        """
        result = "stdout"

        execute_task = self.__progress.add_task("[cyan]Executing job...", total=1)

        w = watch.Watch()

        for event in w.stream(
            self.batch_api_instance.list_namespaced_job,
            namespace=self.namespace,
            label_selector=f"qubernetes.dev/job.type=jupyter",
        ):
            if event["object"].metadata.name == self.name:

                # Job execution completed
                if event["object"].status.active is None:
                    # Failed
                    if event["object"].status.conditions is None:
                        message = "Failed"
                        color = "red"
                        w.stop()
                        result = "stderr"

                    # Succeeded
                    else:
                        message = event["object"].status.conditions[-1].type
                        color = "green"

                        if event["object"].status.conditions[-1].type == "Complete":
                            w.stop()
                # Job schedukled
                elif event["type"] == "ADDED":
                    message = "Scheduled"
                    color = "orange3"
                # Job running
                else:
                    message = "Running"
                    color = "yellow"

                if self.jupyter_logger is None:
                    self.__progress.update(
                        execute_task,
                        description=f"[cyan]Executing job... [{color}]{message}",
                    )
                else:
                    self.jupyter_logger(f"Pod status: {message}")

        self.__progress.advance(execute_task, 1)

        return result

    def __prepare_environment(self):
        """
        Prepare the environment variables.
        """
        env = []

        for key in self.__env.keys():
            env.append(
                client.V1EnvVar(
                    name=key,
                    value_from=client.V1EnvVarSource(
                        secret_key_ref=client.V1SecretKeySelector(
                            name=self.name, key=key
                        )
                    ),
                )
            )

        return env

    def execute(self, code: str) -> tuple[str, str]:
        """
        Execute the given code.
        """

        try:
            self.__create_job_object(code=code)

            if self.jupyter_logger is not None:
                self.jupyter_logger(f"Job {self.name} created")

            stream = self.__complete_and_get_job_status()

            job = self.__get_pods_in_job()
            logs = self.__get_job_logs(job)
            self.__progress.console.print("Fetched job logs")

            return logs, stream
        except KeyboardInterrupt:
            return "Task interrupted by user", "stderr"
        except:
            return "An error occurred.", "stderr"
        finally:
            self.__delete_job()
            pass

    def abort(self):
        """
        Abort the execution.
        """
        self.__delete_job()



import unittest
from unittest.mock import MagicMock, patch
from q8s.kernel import Q8sKernel, kernel_comm_identifier


class TestQ8sKernel(unittest.TestCase):

    @patch("q8s.kernel.CommManager")
    @patch.dict(
        "os.environ",
        {"KUBECONFIG": "./tests/fixtures/config.fixture", "DOCKER_IMAGE": "mock_image"},
    )
    def test_initialize_comm_manager(self, MockCommManager):
        # Mock CommManager instance
        mock_comm_manager = MockCommManager.return_value
        mock_register_target = mock_comm_manager.register_target

        # Create an instance of Q8sKernel
        kernel = Q8sKernel()

        # Assert CommManager was initialized with the kernel instance
        MockCommManager.assert_called_once_with(kernel=kernel)

        # Assert register_target was called with the correct arguments
        mock_register_target.assert_called_once_with(
            kernel_comm_identifier, kernel._on_comm_open
        )

        # Assert shell handlers were set correctly
        self.assertEqual(
            kernel.shell_handlers["comm_open"], mock_comm_manager.comm_open
        )
        self.assertEqual(kernel.shell_handlers["comm_msg"], mock_comm_manager.comm_msg)
        self.assertEqual(
            kernel.shell_handlers["comm_close"], mock_comm_manager.comm_close
        )

    @patch("q8s.kernel.Project")
    @patch("q8s.kernel.Q8sKernel.k8s_context", create=True)
    @patch.dict(
        "os.environ",
        {"KUBECONFIG": "./tests/fixtures/config.fixture", "DOCKER_IMAGE": "mock_image"},
    )
    def test_on_comm_open(self, mock_k8s_context, MockProject):
        # Mock the comm object
        mock_comm = MagicMock()
        mock_comm.comm_id = "mock_comm_id"

        # Mock Project and its configuration
        mock_project_instance = MockProject.return_value
        mock_project_instance.configuration.targets.keys.return_value = [
            "cpu",
            "gpu",
        ]

        # Mock k8s_context target
        mock_k8s_context.target.name = "gpu"

        # Create an instance of Q8sKernel
        kernel = Q8sKernel()

        # Call _on_comm_open
        kernel._on_comm_open(mock_comm, {})

        # Assert comm.send was called with the correct initial message
        mock_comm.send.assert_called_once_with(
            {
                "command": "init",
                "targets": ["cpu", "gpu"],
                "selected_target": "gpu",
            }
        )

    @patch("q8s.kernel.Project")
    @patch("q8s.kernel.Q8sKernel.k8s_context", create=True)
    @patch.dict(
        "os.environ",
        {"KUBECONFIG": "./tests/fixtures/config.fixture", "DOCKER_IMAGE": "mock_image"},
    )
    def test_set_target_comm_msg(self, mock_k8s_context, MockProject):
        # Mock the comm object
        mock_comm = MagicMock()
        mock_comm.comm_id = "mock_comm_id"

        # Mock Project and its configuration
        mock_project_instance = MockProject.return_value
        mock_project_instance.configuration.targets.keys.return_value = [
            "cpu",
            "gpu",
        ]

        # Create an instance of Q8sKernel
        kernel = Q8sKernel()

        kernel.k8s_context = mock_k8s_context

        # Register and invoke the comm_open handler
        # kernel._on_comm_open(mock_comm, {})
        kernel.comm_manager.comm_open(
            stream=None,
            ident=None,
            msg={
                "header": {"msg_type": "comm_open"},
                "content": {
                    "comm_id": mock_comm.comm_id,
                    "target_name": kernel_comm_identifier,
                },
            },
        )

        # Simulate receiving a comm_msg from the frontend
        kernel.comm_manager.comm_msg(
            ident=None,
            stream=None,
            msg={
                "header": {"msg_type": "comm_msg"},
                "content": {
                    "comm_id": mock_comm.comm_id,
                    "data": {
                        "command": "set_target",
                        "payload": {"target": "cpu"},
                    },
                },
            },
        )

        # Assert set_target was called on k8s_context with the correct argument
        mock_k8s_context.set_target.assert_called_once_with("cpu")


if __name__ == "__main__":
    unittest.main()



import os
from ipykernel.kernelbase import Kernel
from ipykernel.comm import CommManager
from rich.progress import Progress, SpinnerColumn, TextColumn
import logging

from q8s.enums import Target
from q8s.execution import K8sContext
from q8s.project import CacheNotBuiltException, Project, ProjectNotFoundException

FORMAT = "[%(levelname)s %(asctime)-15s q8s_kernel] %(message)s"
logging.basicConfig(level=logging.INFO, format=FORMAT)

CODE = ["test_function", "import json; json.dumps(test_function)"]

kernel_comm_identifier = "dev.qubernetes.kernel"


class Q8sKernel(Kernel):
    implementation = "q8s-kernel"
    implementation_version = "0.1.0"
    language = "no-op"
    language_version = "0.1"
    language_info = {
        "name": "Any text",
        "mimetype": "text/x-python",
        "pygments_lexer": "ipython%d" % 3,
        "file_extension": ".py",
        "nbconvert_exporter": "python",
    }
    banner = "q8s"
    comm_manager: CommManager = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.__initialize_comm_manager()

        self.docker_image = os.environ.get("DOCKER_IMAGE", None)
        kubeconfig = os.environ.get("KUBECONFIG", None)

        if kubeconfig is None:
            logging.error("KUBECONFIG not set")
            exit(1)

        if self.docker_image is None:
            logging.error("DOCKER_IMAGE not set")
            exit(1)

        self.k8s_context = K8sContext(
            kubeconfig,
            self.progress,
            Progress(
                SpinnerColumn(),
                TextColumn("[progress.description]{task.description}"),
                expand=True,
            ),
        )
        self.k8s_context.set_container_image(self.docker_image)
        self.k8s_context.set_registry_pat(os.environ.get("REGISTRY_PAT", None))

        logging.info("q8s kernel started")
        logging.info(f"docker image: {self.docker_image}")

    def __initialize_comm_manager(self):
        self.comm_manager = CommManager(
            kernel=self,
        )

        # Register a comm target
        self.comm_manager.register_target(kernel_comm_identifier, self._on_comm_open)

        # Register the comm_open handler
        self.shell_handlers["comm_open"] = self.comm_manager.comm_open
        self.shell_handlers["comm_msg"] = self.comm_manager.comm_msg
        self.shell_handlers["comm_close"] = self.comm_manager.comm_close

    def do_execute(
        self,
        code,
        silent,
        store_history=True,
        user_expressions=None,
        allow_stdin=False,
    ):
        logging.debug(f"Executing code:\n{code}")

        output, stream_name = self.k8s_context.execute(code)

        self.send_response(self.iopub_socket, "clear_output", {"wait": True})

        logging.debug(output)
        logging.debug(stream_name)

        for line in output.split("\n"):
            if line.startswith("data:image/png;base64,"):
                self.send_response(
                    self.iopub_socket,
                    "display_data",
                    {
                        "data": {"image/png": line[22:]},
                        "metadata": {},
                    },
                )
            elif line.startswith("data:image/jpeg;base64,"):
                self.send_response(
                    self.iopub_socket,
                    "display_data",
                    {
                        "data": {"image/jpeg": line[23:]},
                        "metadata": {},
                    },
                )
            elif line.startswith("data:image/svg+xml;base64,"):
                self.send_response(
                    self.iopub_socket,
                    "display_data",
                    {
                        "data": {"image/svg+xml": line[25:]},
                        "metadata": {},
                    },
                )
            else:
                self.send_response(
                    self.iopub_socket,
                    "display_data",
                    {
                        "data": {"text/plain": line},
                        "metadata": {},
                    },
                )

        return {
            "status": "ok",
            # The base class increments the execution count
            "execution_count": self.execution_count,
            "payload": [],
            "user_expressions": {},
        }

    def progress(self, msg):
        self.send_response(
            self.iopub_socket,
            "display_data",
            {
                "data": {"text/plain": msg},
                "metadata": {},
            },
        )

    def _on_comm_open(self, comm, msg):
        """Handle the frontend opening a Comm"""
        logging.info("Comm opened by frontend")
        logging.info(f"Comm ID: {comm.comm_id}")

        # Send a message to the frontend with the current state of the kernel
        comm.send(
            {
                "command": "init",
                "targets": Project().configuration.targets.keys(),
                "selected_target": self.k8s_context.target.name,
            }
        )

        @comm.on_msg
        def _recv(msg):
            data = msg["content"]["data"]
            logging.info(f"Received from frontend: {data}")

            # Respond to the frontend
            if data["command"] == "set_target":
                self.k8s_context.set_target(Target(data["payload"]["target"]))
                project = Project()
                image = project.cached_images(data["payload"]["target"])
                self.k8s_context.set_container_image(image)
                logging.info(f"Updated execution target to {data['payload']['target']}")
            else:
                logging.warning(f"Unknown command: {data['command']}")

        @comm.on_close
        def _closed(msg):
            logging.info("Comm closed by frontend")
